///////////////////////////////////////////////////////////////////////////////////////////////////
// Hierarchical Saliency Detection Code
//
// It implements the saliency detection algorithm described in the following paper:
//  [1] "Hierarchical Saliency Detection", Qiong Yan, Li Xu, Jianping Shi, Jiaya Jia.
//      IEEE Conference on Computer Vision and Pattern Recognition (CVPR), 2013 
//  
// The method is mainly implemented in this function
//  void MultiscaleSaliency(const char *in_file, const char *out_dir, int *min_reg, int *scales, 
//			int sigma_s, float sigma_r, int noSmooth, float bdWeight, float *beta, float *lambda);
//
// Author  : Qiong Yan (june232.ustc@gmail.com)
// Date    : 04/12/2015
// Version : 1.0 
// Copyright 2015, The Chinese University of Hong Kong.
///////////////////////////////////////////////////////////////////////////////////////////////////

#include <vector>
#include <fstream>
#include <iostream>
#include <cstdio>
#include <cstdlib>
#include <string>
#include <algorithm>
#include <windows.h>
#include <tchar.h>
#include <time.h>
#include "opencv2/core/core.hpp"
#include "opencv2/highgui/highgui.hpp"
#include "opencv2/imgproc/imgproc.hpp"
#include "Segmentation.h"
#include "SegTree.h"
#include "DomainTransform.h"
#include "MyTime.h"


//#define GENERATE_CHS_DATA_FOR_INFERENCE		// comment this line to generate HS result
#define REMOVE_IMAGE_FRAME
#define AVERAGE_TIME
#define SALIENCY_MODEL_L2
//#define IMAGE_DOWNSAMPLE
//#define SAVE_LAYER_RESULT

#define ARGC_MIN               3
#define ARGC_SCALES_ALL        4
#define IMAGE_DOWNSAMPLE_RATIO (0.5)
#define SALIENT_RATIO          (0.01f)
#define UNSALIENT_RATIO        (0.01f)
#define SQUARE(x)		       ((x)*(x))	
#define PI					   (3.14159265)
#define PI_SQRT				   (1.77245385)
#define	Loop_Image(x,y,width,height) for(int y=0; y<height; y++) for(int x=0; x<width; x++)


#ifdef GENERATE_CHS_DATA_FOR_INFERENCE
#include "mat.h"
#endif


using namespace std;
using namespace cv;


struct SegSizeSal
{
	float sal, sz;
	int id;

	bool operator < (SegSizeSal &s2)
	{
		return sal<s2.sal;
	}
};

///////////////////////////////////////////////////////////////////////////////////////////////////
// some constants
///////////////////////////////////////////////////////////////////////////////////////////////////
const int N_SCALES = 3;
const int MAX_NUM_COLOR = 1500;
const int debug = 0;





///////////////////////////////////////////////////////////////////////////////////////////////////
// timing
///////////////////////////////////////////////////////////////////////////////////////////////////
clock_t cCheckTime = 0;
clock_t cummulateTime = 0;
clock_t cummulateCounter = 0;

void CCLEAR(){
	cummulateTime = 0;
	cummulateCounter = 0;
}

void CSTART(){
	cCheckTime = clock();
}

void CEND(){
	clock_t newTime = clock();
	cummulateTime += newTime - cCheckTime;
	cCheckTime = newTime;

	cummulateCounter++;
}

void CSUMMARY(){
	printf("cummulate time: %d ms\n",cummulateTime);
	//printf("cummulate counter: %d\n",cummulateCounter);
}



///////////////////////////////////////////////////////////////////////////////////////////////////
// New on August 1, 2012 by Yan Qiong
// Get all files specified by fname (single or multiple)
///////////////////////////////////////////////////////////////////////////////////////////////////
int GetFiles(char *fname, vector<string> &files)
{
	files.clear();
	WIN32_FIND_DATAA fileFindData;
	HANDLE hFind = FindFirstFileA(fname, &fileFindData);
	if (hFind == INVALID_HANDLE_VALUE)
		return 0;

	do 
	{
		if (fileFindData.cFileName[0] == '.')
			continue; // filter the '..' and '.' in the path
		if (fileFindData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
			continue; // Ignore sub-folders
		files.push_back(fileFindData.cFileName);
	} while (FindNextFileA(hFind, &fileFindData));
	FindClose(hFind);
	return (int)files.size();
}


// Extract file name without parent path and extension
int ExtractFileName(const char *in_file, char *in_file_no_ext)
{
	strcpy(in_file_no_ext, in_file);
	int path_len = strlen(in_file_no_ext);
	int id = path_len;

	// find and remove path
	while(id--) if(in_file_no_ext[id]=='\\' || in_file_no_ext[id]=='/') break;
	if( id==0 && ( in_file_no_ext[id]!='\\'|| in_file_no_ext[id]!='/' ) ) id = -1;
	strcpy(in_file_no_ext, in_file_no_ext+id+1);

	// find and remove extension
	id = strlen(in_file_no_ext);
	while(id--) if(in_file_no_ext[id]=='.') break;
	in_file_no_ext[id] = '\0';

	return strlen(in_file_no_ext);
}



// Make at 1% pixel un-salient and 1% salient
void SalScale(float *segSal, float *segSz, int n)
{
	//if( n<4 ) return;

	SegSizeSal *seg = new SegSizeSal[n];

	for (int i=0; i<n; i++)
	{
		seg[i].id = i;
		seg[i].sal = segSal[i];
		seg[i].sz = segSz[i];
	}

	sort(seg, seg+n);

	int max_id, min_id;
	float low_sum, high_sum;

	low_sum = 0;
	for (min_id=0; min_id<n, low_sum<UNSALIENT_RATIO; min_id++) low_sum += seg[min_id].sz;

	high_sum = 0;
	for (max_id=n-1; max_id>=0, high_sum<SALIENT_RATIO; max_id--) high_sum += seg[max_id].sz;
	
	min_id = max(min_id-1, 0); max_id = min(max_id+1, n-1);

	low_sum = seg[min_id].sal;
	high_sum = seg[max_id].sal;
	for (int i=0; i<n; i++) 
	{
		segSal[i] = 255.0f * max(0.0f, min(1.0f, (segSal[i] - low_sum) / (high_sum - low_sum))); 
	}
	
	delete []seg;
}


//////////////////////////////////////////////////////////////////////
// Remove image frame
//////////////////////////////////////////////////////////////////////
void RemoveFrame(Mat &in, Mat &out, int &top, int &bot, int &lef, int &rig)
{
	double edge_th = 75; 
	Mat gray, edges;
	//in.convertTo(gray, CV_BGR2GRAY);
	cvtColor(in, gray, CV_BGR2GRAY);
	Canny(in, edges, edge_th*0.4, edge_th);

	int rows = in.rows, cols = in.cols;
	int max_width = in.rows/8;
	double th = 0.6*255;
	int flag_top, flag_bot, flag_lef, flag_rig;
	top = 0;
	bot = 0;
	lef = 0;
	rig = 0;
	for (int k=1; k<max_width; k++)
	{
		if ( mean(edges.row(k-1))[0] > th )
		{
			flag_top = 1;
			top = k;
		}
		if ( mean(edges.row(rows-k))[0] > th )
		{
			flag_bot = 1;
			bot = k;
		}
		if ( mean(edges.col(k-1))[0] > th )
		{
			flag_lef = 1;
			lef = k;
		}
		if ( mean(edges.col(cols-k))[0] > th )
		{
			flag_rig = 1;
			rig = k;
		}
	}

	if ( flag_top + flag_bot + flag_lef + flag_rig > 1 )
	{
		max_width = top;
		if (bot > max_width) max_width = bot;
		if (lef > max_width) max_width = lef;
		if (rig > max_width) max_width = rig;

		if (top==0) top = max_width;
		if (bot==0) bot = max_width;
		if (lef==0) lef = max_width;
		if (rig==0) rig = max_width;

		out = in.rowRange(top, rows-bot).colRange(lef, cols-rig).clone();
	}
}



void ClearSegList(vector<SegPatch*> &seg_list)
{
	if(seg_list.size()>0){
		for(int i=0;i<seg_list.size();i++)
		{
			(seg_list[i])->Clear();
			delete seg_list[i];
		}
		seg_list.clear();
	}
}



void CreateClrList(Mat &seg_res, vector<SegPatch*>&seg_list, uchar *clr_list_rgb)
{
	int x, y;
	uchar *u_ptr;

	for (int i=0; i<seg_list.size(); i++)
	{
		x = (seg_list[i]->m_ptlist[0]).x;
		y = (seg_list[i]->m_ptlist[0]).y;
		u_ptr = seg_res.data + (y*seg_res.cols+x)*3;
		clr_list_rgb[i*3+0] = u_ptr[2];
		clr_list_rgb[i*3+1] = u_ptr[1];
		clr_list_rgb[i*3+2] = u_ptr[0];
	}
}


void CreateSegList(Mat& segImg, vector<SegPatch*>& seg_list)
{
	int i,j;
	int iMaxSegIndex = 0;
	int iWidth = segImg.cols;
	int iHeight = segImg.rows;
	int *i_ptr;

	ClearSegList(seg_list);

	i_ptr = segImg.ptr<int>();
	for (j=0; j<iHeight*iWidth; j++)
		iMaxSegIndex = max(iMaxSegIndex, i_ptr[j]);
	
	seg_list.resize(iMaxSegIndex+1);
	for(i=0;i<=iMaxSegIndex;i++){
		seg_list[i] = new SegPatch();
		seg_list[i]->m_iLabel = i;
		seg_list[i]->Clear();
		seg_list[i]->ox = 0;
		seg_list[i]->oy = 0;
	}

	int label;
	i_ptr = segImg.ptr<int>();
	for (j=0; j<iHeight; j++)
	{
		for (i=0; i<iWidth; i++)
		{
			label = i_ptr[j*iWidth+i];
			seg_list[label]->Add(i,j);
			seg_list[label]->ox += i;
			seg_list[label]->oy += j;
		}
	}

	for(i=0; i<=iMaxSegIndex; i++)
	{
		seg_list[i]->ox /= (float)(seg_list[i]->Size());
		seg_list[i]->oy /= (float)(seg_list[i]->Size());
	}

	// remove null seg
	vector<SegPatch*>::iterator it = seg_list.begin();
	int n = seg_list.size();
	for(i=0;i<n;i++)
	{
		if((*it)->Size() != 0)
			it++;
		else
			seg_list.erase(it);
	}


	//CheckSegment(100,100);
	for(i=0;i<seg_list.size();i++){
		seg_list[i]->CalBoardList(segImg);
	}
	//update segmap (remove empty labels & make sure pixel labels are the same with its order in the list
	i_ptr = segImg.ptr<int>();
	for(i=0;i<seg_list.size();i++){
		std::vector<Point>::iterator iter = seg_list[i]->m_ptlist.begin();	
		for(;iter!=seg_list[i]->m_ptlist.end();iter++){			
			i_ptr[(iter->y)*iWidth+iter->x] = i;
		}
		seg_list[i]->m_iLabel = i;
	}
}

////////////////////////////////////////////////////////////////////////////////////////////////////
//// Save segmentation list
////////////////////////////////////////////////////////////////////////////////////////////////////
//void SaveSegList(char *out_name, Mat *seg_img, vector<SegPatch*>* seg_list, float **sal, int *scales, int N_SCALES)
//{
//	FILE *txt = fopen(out_name, "w");
//	if( txt==NULL )
//	{
//		printf("Cannot open file for writing!\n");
//		return;
//	}
//
//	fprintf(txt, "Total Scales: %d\n", N_SCALES);
//
//	for( int k=0; k<N_SCALES; k++ )
//	{
//		int n_segs = seg_list[k].size();
//		fprintf(txt, "Scale: %d Segment %d\n", scales[k], n_segs);
//		
//		vector<SegPatch*> &segs = seg_list[k];
//		for( int i=0; i<n_segs; i++ )
//		{
//			vector<Point> pixels = segs[i]->m_ptlist;
//			uchar *color =  &(seg_img[k].at<uchar>(pixels[0].y, pixels[0].x, 0));
//
//			fprintf(txt, "Segment %d Color %d %d %d Saliency %f\n", i+1, color[2], color[1], color[0], sal[k][i]);
//
//			for( int p=0; p<pixels.size(); p++)
//				fprintf(txt, "(%d, %d) ", pixels[p].x, pixels[p].y);
//
//			fprintf(txt, "\n\n");
//		}
//	}
//
//	fclose(txt);
//}


#ifdef GENERATE_CHS_DATA_FOR_INFERENCE
//////////////////////////////////////////////////////////////////////////////////////////////////
// Save segmentation list into Matlab .mat file
//////////////////////////////////////////////////////////////////////////////////////////////////
int SaveSegList2(char *out_name, Mat *seg_labels, Mat *seg_images, vector<SegPatch*>* seg_list, float **sal, int *scales, int N_SCALES, int *border)
{
	// Open a file for saving Matlab variables
	MATFile *mat_file = matOpen(out_name, "w");
	if( mat_file == NULL ){
		printf("Error reopening file %s\n", out_name);
		return -1;
	}

	// Define Matlab variable
	mxArray *mat;
	int ndim;
	mwSize *dimsize;

	mat = mxCreateDoubleScalar(N_SCALES);
	matPutVariable(mat_file, "n_scales", mat);
	mxDestroyArray(mat);

	// save the scales
	// Allocate memory
	mat = mxCreateDoubleMatrix(N_SCALES, 1, mxREAL );
	// set data
	double *mat_data = mxGetPr(mat);
	for(int k=0; k<N_SCALES; k++) mat_data[k] = scales[k];
	// write data to file
	matPutVariable(mat_file, "scales", mat);
	// Deallocate memory
	mxDestroyArray(mat);

	// save seg numbers
	mat = mxCreateDoubleMatrix(N_SCALES, 1, mxREAL);
	mat_data = mxGetPr(mat);
	for(int k=0; k<N_SCALES; k++) mat_data[k] = seg_list[k].size();
	matPutVariable(mat_file, "n_segs", mat);
	mxDestroyArray(mat);

	// save seg labels (the label image)
	int rows = seg_labels[0].rows;
	int cols = seg_labels[0].cols;
	ndim = 3;
	dimsize = new mwSize[ndim];
	dimsize[0] = rows;
	dimsize[1] = cols;
	dimsize[2] = N_SCALES;
	mat = mxCreateNumericArray(ndim, dimsize, mxDOUBLE_CLASS, mxREAL);
	mat_data = mxGetPr(mat);
	for(int k=0; k<N_SCALES; k++)
	{
		int *img_data = seg_labels[k].ptr<int>(0);
		for(int i=0; i<cols; i++)
			for(int j=0; j<rows; j++)
				mat_data[(k*cols+i)*rows+j] = img_data[j*cols+i];
	}
	matPutVariable(mat_file, "seg_labels", mat);
	mxDestroyArray(mat);
	delete []dimsize;

	// save seg color
	int n_seg0 = seg_list[0].size();
	ndim = 3;
	dimsize = new mwSize[ndim];
	dimsize[0] = 3;
	dimsize[1] = n_seg0;
	dimsize[2] = N_SCALES;
	mat = mxCreateNumericArray(ndim, dimsize, mxDOUBLE_CLASS, mxREAL);
	mat_data = mxGetPr(mat);
	for(int k=0; k<N_SCALES; k++)
	{
	 	uchar *img_data = seg_images[k].ptr<uchar>(0);
	 	//double r = 1 / 255.0;
	 	int sz = k*n_seg0*3;
		int n_seg = seg_list[k].size();
		for(int i=0; i<n_seg; i++)
		{
			int x = (seg_list[k])[i]->m_ptlist[0].x;
			int y = (seg_list[k])[i]->m_ptlist[0].y;
			mat_data[sz+i*3] = img_data[(y*cols+x)*3+2];
			mat_data[sz+i*3+1] = img_data[(y*cols+x)*3+1];
			mat_data[sz+i*3+2] = img_data[(y*cols+x)*3];
		}
	}
	matPutVariable(mat_file, "seg_colors", mat);
	mxDestroyArray(mat);
	delete []dimsize;

	// save seg size (number of pixels of each segment)
	mat = mxCreateDoubleMatrix(n_seg0, N_SCALES, mxREAL);
	mat_data = mxGetPr(mat);
	for(int k=0; k<N_SCALES; k++)
		for(int i=0; i<seg_list[k].size(); i++)
			mat_data[k*n_seg0+i] = (seg_list[k])[i]->m_ptlist.size();
	matPutVariable(mat_file, "seg_sizes", mat);
	mxDestroyArray(mat);

	// border
	mat = mxCreateDoubleMatrix(4, 1, mxREAL);
	mat_data = mxGetPr(mat);
	if ( border==NULL )
	{
		mat_data[0] = mat_data[1] = mat_data[2] = mat_data[3] = 0;
	}
	else
	{
		mat_data[0] = border[0];
		mat_data[1] = border[1];
		mat_data[2] = border[2];
		mat_data[3] = border[3];
	}
	matPutVariable(mat_file, "border", mat);
	mxDestroyArray(mat);

// 	// save seg image (the color image)
// 	ndim = 4;
// 	dimsize = new mwSize[ndim];
// 	dimsize[0] = rows;
// 	dimsize[1] = cols;
// 	dimsize[2] = 3;
// 	dimsize[3] = N_SCALES;
// 	mat = mxCreateNumericArray(ndim, dimsize, mxDOUBLE_CLASS, mxREAL);
// 	mat_data = mxGetPr(mat);
// 	for(int k=0; k<N_SCALES; k++)
// 	{
// 		uchar *img_data = seg_images[k].ptr<uchar>(0);
// 		//double r = 1 / 255.0;
// 		int sz = k*rows*cols*3;
// 		for(int c=0; c<3; c++)
// 			for(int i=0; i<cols; i++)
// 				for(int j=0; j<rows; j++)
// 					mat_data[sz+(c*cols+i)*rows+j] = img_data[(j*cols+i)*3+2-c];
// 	}
// 	matPutVariable(mat_file, "seg_images", mat);
// 	mxDestroyArray(mat);
// 	delete []dimsize;
	
	// save saliency values
	mat = mxCreateDoubleMatrix(n_seg0,N_SCALES, mxREAL);
	mat_data = mxGetPr(mat);
	for(int k=0; k<N_SCALES; k++)
		for (int i=0; i<seg_list[k].size(); i++)
			mat_data[k*n_seg0 + i] = sal[k][i];
	matPutVariable(mat_file, "saliency", mat);
	mxDestroyArray(mat);

	// Close the Matlab file
	matClose(mat_file);
	return 1;
}
#endif

//////////////////////////////////////////////////////////////////////////////////////////////////
// lambda controls the cross-layer smoothness
// beta penalizes the deviation from initial value
//////////////////////////////////////////////////////////////////////////////////////////////////
void SumProductOnTree(SegTree *tree, float *salFused, int N_SCALES, float *lambda, float *beta)
{
	// now only deal with the case of 3 scales
	//if (N_SCALES==3)
//	{
	
	SegTreeNode *p = tree->root; 
	int i;

	for (int i=0; i<tree->n_nodes; ++i, ++p)
	{
		p->value_acc = 0;
		p->weight_acc = 0;
	}
	
	int level, lambdai, betai;

	float WH_ = 1.0f / tree->root->segmt->SizeSq();

	//////////////////////////////////////////////////////////////////////////////////////////////////
	// Bottom-up pass
	//////////////////////////////////////////////////////////////////////////////////////////////////

	p = tree->root + tree->n_nodes - 1;
	for ( ; p->level<N_SCALES; --p)						// loop until the root
	{
		lambdai = lambda[p->level];
		betai = beta[p->level] ;//* max(1, p->nchild);

		// message to the upper layer
		p->value_up = betai * p->value0 + p->value_acc;		
		p->weight_up = lambdai + betai + p->weight_acc;
		
		p->parnt->value_acc += p->value_up * lambdai / p->weight_up;				
		p->parnt->weight_acc += ( p->weight_up - lambdai ) * lambdai / p->weight_up;
	}

	// calculate value of the top layer
	p=tree->root;
	p->value1 = ( beta[N_SCALES] /** max(1, p->nchild)*/ * p->value0 + p->value_acc ) / ( beta[N_SCALES] /** max(1, p->nchild)*/ + p->weight_acc );


	//////////////////////////////////////////////////////////////////////////////////////////////////
	// Top-down pass
	//////////////////////////////////////////////////////////////////////////////////////////////////

	i = 1; 	p++;
	for ( ; i<tree->n_nodes; i++, p++)
	{
		lambdai = lambda[p->level];

		p->value1 = ( lambdai * p->parnt->value1 + p->value_up ) / p->weight_up;
	}

	p = tree->root + tree->n_nodes - 1;
	for ( ; p->level==0; p--)
	{
		salFused[p->segmt->m_iLabel] = p->value1;
	}
	
//	}
}

void AutoAdjust(Mat &src, Mat &dst)
{
	int H[256], Ha[256], n;

	memset(H, 0, sizeof(int)*256);

	uchar *p, *q;
	
	p = src.data;
	n = src.rows*src.cols*src.channels();
	for (int i=0; i<n; i++) H[p[i]]++;

	int v1, v2;
	double sum = 0;

	n = n*0.005;
	for (v1=0, sum=0; v1<256, sum<n; v1++) sum += H[v1];
	if( v1 ) v1--;

	for (v2=255, sum=0; v2>0, sum<n; v2--) sum += H[v2];
	if( v2<255 ) v2++;

	
	for (int i=0; i<256; i++)
		Ha[i] = max(0.0f, min(255.0f, (float)((i-v1))/(v2-v1) * 255.0f + 0.5f));

	if (dst.data==NULL || dst.rows!=src.rows || dst.cols!=src.cols)
		dst.create(src.rows, src.cols, src.type());

	p = src.data;
	q = dst.data;
	n = src.rows * src.cols * src.channels();
	for (int i=0; i<n; i++) q[i] = Ha[ p[i] ];
}



///////////////////////////////////////////////////////////////////////////////////////////////
//
// Multi-scale Saliency Detection
//
// Parameters:
//		in_files	- input image file
//		out_dir		- output directory
//		min_reg		- minimum region size for each layer (not used)
//		scales		- scale thresholds for each layer (see our paper for more details)
//		sigma_s		- (not used)
//		sigma_r		- range sigma for segmentation (0~255, default is 6.5)
//		noSmooth	- 0 (default): apply smooth filtering to the input; 1: otherwise
//		bdWeight	- (not used)
//		beta		- single layer saliency confidence weight (see our paper for more details)
//		lambda		- cross layer saliency consistence weight (see our paper for more details)
//
///////////////////////////////////////////////////////////////////////////////////////////////
void MultiscaleSaliency(const char *in_file, const char *out_dir, int *min_reg, int *scales, 
			int sigma_s, float sigma_r, int noSmooth, float bdWeight, float *beta, float *lambda)
{	
	///////////////////////////////////////////////////////////////////////////////////////////////
	//
	// setup input & output
	//
	///////////////////////////////////////////////////////////////////////////////////////////////

	Mat src = imread(in_file);							// uchar 3-channel rgb image
	if (src.empty())
	{
		printf("Cannot open file \"%s\"\nExiting...\n", in_file);
		return;
	}
	if (src.channels() != 3)
	{
		printf("Cannot process gray-scale image\nExiting..\n");
		return;
	}

#ifdef REMOVE_IMAGE_FRAME
	// Remove image frame
	Mat src_backup = src.clone();
	int top, bot, lef, rig;
	RemoveFrame(src_backup, src, top, bot, lef, rig);
#endif

	// get file name of input (without extension)
	char buffer[1024], in_file_no_ext[1024];
	ExtractFileName(in_file, in_file_no_ext);


#ifdef IMAGE_DOWNSAMPLE
	Mat tmp_mat;
	int ori_width = src.cols;
	int ori_height = src.rows;
	resize(src, src, Size(ori_width*IMAGE_DOWNSAMPLE_RATIO, ori_height*IMAGE_DOWNSAMPLE_RATIO));
#endif

	int width = src.cols;								// source image dimension
	int height = src.rows;
	int WH = width*height;
	
	Mat src_filtered(height, width, CV_8UC3);			// uchar 3-channel rgb image
	Mat seg_ress[N_SCALES];								// uchar 3-channel rgb image (segmentation result)
	Mat sal_maps[N_SCALES];								// uchar 1-channel image (output saliency map)
	Mat seg_labels[N_SCALES];							// int 1-channel image (segmentation labels)

	Mat gaussian_map(height, width, CV_32F);			// Gaussian weight map for fast compute center prior
	float *cent_wt = NULL, *size_wt = NULL;				// center weight and relative size
	float *loca_prior = NULL;							// final background prior

	std::vector<SegPatch*> seg_lists[N_SCALES];			// segmentation list for each scale
	float *sal[N_SCALES] = {0}, *sal_merged = NULL;		// saliency at each layer and the final fused
	float *clr_distance = NULL, *clr_contrast = NULL;	// color distance, boundary distance, color contrast
	float *hist = NULL, *hist_sqrt = NULL;				// histogram (float)
	int *ihist_bd = NULL;								// boundary histogram (int)
	Mat clr_rgb, clr_lab;
	uchar *clr_lists_rgb[N_SCALES];						// color list, one for a region
	float *sal_tmp = NULL, *wt_tmp = NULL;

	
	float clr1[3], clr2[3];
	int n_segs;

	uchar *u_ptr;
	int *i_ptr;
	

#ifdef AVERAGE_TIME
	CSTART();
#endif


	///////////////////////////////////////////////////////////////////////////////////////////////
	//
	// Initialization step (memory allocation, and some other pre-calculations)
	//
	///////////////////////////////////////////////////////////////////////////////////////////////


	// Allocate memory
	for( int i=0; i<N_SCALES; i++ )
	{
		seg_labels[i].create(height, width, CV_32S);
		seg_ress[i].create(height, width, CV_8UC3);
		sal_maps[i].create(height, width, CV_8U);
	}


	// Init gaussian_map (a map the same size as input and it looks like gaussian kernel )
	// for center prior
	Mat Gau_x, Gau_y;
	Gau_x = getGaussianKernel((width+1)/2*2+1, width/3.0, CV_32F);
	Gau_y = getGaussianKernel((height+1)/2*2+1, height/3.0, CV_32F);

	float *f_ptr_x, *f_ptr_y, *f_ptr;
	f_ptr_x = Gau_x.ptr<float>();
	f_ptr_y = Gau_y.ptr<float>();
	f_ptr = gaussian_map.ptr<float>();

	int cx = (width+1)/2;
	int cy = (height+1)/2;
	for ( int i=0; i<Gau_x.rows; i++ ) 	
	{
		f_ptr_x[i] = min(1.0f, f_ptr_x[i]/(0.8f*f_ptr_x[cx]));
	}
	for ( int i=0; i<Gau_y.rows; i++ )
	{
		f_ptr_y[i] = min(1.0f, f_ptr_y[i]/(0.8f*f_ptr_y[cy]));
	}

	Loop_Image(x,y,cx,cy)
	{
		f_ptr[ y*width+x ] = f_ptr[ y*width+(width-1-x) ] 
						   = f_ptr[ (height-1-y)*width+x ] 
						   = f_ptr[ (height-1-y)*width+(width-1-x) ] 
						   = f_ptr_x[ x ] * f_ptr_y[ y ];
	}

	for ( int y=0; y<height; y++ )
	{
		f_ptr[y*width] = f_ptr[(y+1)*width-1] = -10;
	}
	for ( int x=0; x<width; x++ )	
	{
		f_ptr[x] = f_ptr[(height-1)*width+x] = -10;
	}
	//imwrite("GaussianMap.png", gaussian_map*255);

	
#ifdef SALIENCY_TIMER_ALL
	clock_t start, end, start0;	
	start0 = clock();												// clock for the whole running stage
#endif





	///////////////////////////////////////////////////////////////////////////////////////////////
	//
	// Hierarchical (Multi-scale) Saliency Detection
	// It contains four major step: 
	//		1 - pre-smoothing
	//		2 - Layer extraction by segmentation
	//		3 - Single layer saliency estimation
	//		4 - Saliency Merge
	//
	///////////////////////////////////////////////////////////////////////////////////////////////




	///////////////////////////////////////////////////////////////////////////////////////////////
	//
	// Stage 1: Pre-smoothing using smoothing filter (bilateral filter or recursive filter)
	//          to remove tiny details for better segmentation
	//
	// if 'noSmooth', we use 'src' image as input for the next step
	//
	///////////////////////////////////////////////////////////////////////////////////////////////


#ifdef SALIENCY_TIMER_ALL
	start = clock();
#endif

	// Scale the input image range to full range 0~255
	AutoAdjust(src, src);

	float rf_sigma_s = 12.0f, rf_sigma_r = 0.08f;

	if( noSmooth )
	{
		src_filtered = src;
	}
	else
	{
		src.convertTo(src_filtered, CV_32F);//, 1/255.0);
		RF(src_filtered, src_filtered, rf_sigma_s*255, rf_sigma_r*255);
		src_filtered.convertTo(src_filtered, CV_8U);//, 255.0);
	}


#ifdef SALIENCY_TIMER_ALL
	end = clock();
	printf("Stage 1: Bilateral filtering (%.6fs)\n", (double)(end-start)/CLOCKS_PER_SEC);
#endif




	///////////////////////////////////////////////////////////////////////////////////////////////
	//
	// Stage 2: Layer extraction by segmentation via region grow and merge
	//
	// input image is 'src_filtered', minimum region size set by 'min_reg', scale parameter by 'scales', 
	// number of scales by 'N_SCALES', 'sigma_s', 'sigma_r' are the spatial sigma and range sigma
	// parameters, respectively. (min_reg, sigma_s are not used)
	//
	// this step will give 3 segmentation results, region labels stored in 'seg_labels', segmentation
	// stored in 'seg_ress', list of segments in 'seg_lists', color of each segment in 'clr_lists_rgb' 
	// color of each segment is assigned as the dominant color in this region
	//
	// E.g.
	// seg_labels[0]      - (CV_32SC1) Segmentation labels for the first layer. It has the same size 
	//                      as the input image. Value of each pixel represents which region belongs to.
	// seg_ress[0]        - (CV_8UC3) Segmentation result for the first layer. It has the same size 
	//                      as the input image. Value of each pixel is its color after segmentation.
	// clr_lists_rgb[0]   - (uchar array) Stores all existing colors in the first image layer.
	// seg_lists[0]       - (vector of SegPatch) Stores all segments in the first image layer.
	///////////////////////////////////////////////////////////////////////////////////////////////

#ifdef SALIENCY_TIMER_ALL
	start = clock();
#endif
	
	SegmentImage(&src_filtered, seg_labels, seg_ress, min_reg, scales, N_SCALES, sigma_s, sigma_r);
	for (int ii=0; ii<N_SCALES; ii++)
	{
		CreateSegList(seg_labels[ii], seg_lists[ii]);					// seg label is same as its index in the list
																		// (seg_lists[ii])[j]->m_iLabel == j		
		clr_lists_rgb[ii] = new uchar[seg_lists[ii].size()*3];
		CreateClrList(seg_ress[ii], seg_lists[ii], clr_lists_rgb[ii]);	// clr_lists_rgb has the same order as seg_lists
	}

#ifdef SALIENCY_TIMER_ALL
	end = clock();
	printf("Stage 2: Mean-shift segmentation (%.6fs)\n", (double)(end-start)/CLOCKS_PER_SEC);
#endif





	///////////////////////////////////////////////////////////////////////////////////////////////
	//
	// Stage 3: Single layer saliency estimation given regions from previous step
	//
	// First, region contrast is computed using their color difference
	// Secondly, we apply location prior to each region
	// Thirdly, map saliency of each region, result of each scales is saved (not now)
	// Finally, merge saliency of the three scales, save the final result
	//
	///////////////////////////////////////////////////////////////////////////////////////////////


#ifdef SALIENCY_TIMER_ALL
	start = clock();
#endif

	n_segs		= seg_lists[0].size(); 
	hist		= new float[n_segs*N_SCALES];
	hist_sqrt	= new float[n_segs*N_SCALES];
	ihist_bd	= new int[n_segs*N_SCALES];
	clr_distance= new float[n_segs*n_segs];
	clr_contrast= new float[n_segs*n_segs];
	loca_prior	= new float[n_segs*N_SCALES];
	cent_wt		= new float[n_segs*N_SCALES];
	size_wt		= new float[n_segs*N_SCALES];
	sal_tmp		= new float[n_segs];
	wt_tmp		= new float[n_segs];


	for( int k=0; k<N_SCALES; k++ )
	{

		n_segs = seg_lists[k].size();

		///////////////////////////////////////////////////////////////////////////////////////////
		//
		// Step 3.1: calculate region distance using their color difference in Lab space
		//           colors of each region is stored in 'clr_lists_rgb' in RGB space
		//
		///////////////////////////////////////////////////////////////////////////////////////////
		
		float sigma_s_contrast = min(0.5f, 2*0.04f*scales[k]*0.5f);

		clr_rgb.create(n_segs, 1, CV_8UC3);
		clr_lab.create(n_segs, 1, CV_8UC3);
		memcpy( clr_rgb.data, clr_lists_rgb[k], n_segs*3*sizeof(uchar));
		cvtColor(clr_rgb, clr_lab, CV_RGB2Lab);

		memset(clr_distance, 0, n_segs*n_segs*sizeof(float));
		memset(clr_contrast, 0, n_segs*n_segs*sizeof(float));
		
		u_ptr = clr_lab.data;

		for( int i=0; i<n_segs; i++ )
		{
			clr1[0] = u_ptr[i*3+0];
			clr1[1] = u_ptr[i*3+1];
			clr1[2] = u_ptr[i*3+2];

			float wij = 0;		
			for( int j=i+1; j<n_segs; j++ )
			{
				clr2[0] = u_ptr[j*3+0];
				clr2[1] = u_ptr[j*3+1];
				clr2[2] = u_ptr[j*3+2];
				
				clr_distance[j*n_segs+i] = min(SQUARE(320.0f), ( (clr1[0]-clr2[0]) * (clr1[0]-clr2[0]) + 8 * 
					( (clr1[1]-clr2[1])*(clr1[1]-clr2[1])+ (clr1[2]-clr2[2])* (clr1[2]-clr2[2]))) );

				wij = exp( - ((seg_lists[k])[i])->TransformedDistance((seg_lists[k])[j], width, height)/(sigma_s_contrast));

				clr_contrast[j*n_segs+i] = sqrt(clr_distance[j*n_segs+i]) * wij;
				clr_distance[i*n_segs+j] = clr_distance[j*n_segs+i];
				clr_contrast[i*n_segs+j] = clr_contrast[j*n_segs+i];
			}
		}




		///////////////////////////////////////////////////////////////////////////////////////////
		//
		// Step 3.2: calculate histogram (float, normalized), i.e. region size
		//			 hist[i] is also the normalized size of region i.
		//
		///////////////////////////////////////////////////////////////////////////////////////////
		
		for( int i=0; i<n_segs; i++ ) 
		{
			hist[i] = ((seg_lists[k])[i])->Size() / (float)(WH); 
			hist_sqrt[i] = sqrt(hist[i]);
		}

		///////////////////////////////////////////////////////////////////////////////////////////
		//
		// Step: 3.3 calculate location prior for each region
		//
		///////////////////////////////////////////////////////////////////////////////////////////
		
		memset(cent_wt, 0, n_segs*sizeof(float));

		i_ptr = seg_labels[k].ptr<int>();
		f_ptr = gaussian_map.ptr<float>();
		for ( int y=0; y<WH; y++ )
		{
			cent_wt[ i_ptr[y] ] += f_ptr[y];
		}

		for ( int i=0; i<n_segs; i++ )
		{

			// center weight: sum of gaussian distance to the image center
			cent_wt[i] = max(0.0f, cent_wt[i] / (seg_lists[k])[i]->Size() );


			loca_prior[i] = cent_wt[i];
 			if( k>0 && loca_prior[i] < 0.5 )	// only apply to the middle and top layer
			{	
				loca_prior[i] *= exp( min(0.0f, 2 - sqrt(hist[i]*n_segs) ) ); 
			}

		}


		///////////////////////////////////////////////////////////////////////////////////////////
		//
		// Step: 3.3: calculate boundary histogram, boundary width is 2 pixel
		//
		///////////////////////////////////////////////////////////////////////////////////////////

		i_ptr = seg_labels[k].ptr<int>();
		memset(ihist_bd, 0, n_segs*sizeof(int));	
		for( int x=0; x<width; x++ )	
		{
			ihist_bd[ i_ptr[x] ] ++;
			ihist_bd[ i_ptr[width+x] ] ++;
			ihist_bd[ i_ptr[(height-2)*width+x] ] ++;
			ihist_bd[ i_ptr[(height-1)*width+x] ] ++;
		}
		for( int y=0; y<height; y++ )
		{
			ihist_bd[ i_ptr[y*width] ] ++;
			ihist_bd[ i_ptr[y*width+1] ] ++;
			ihist_bd[ i_ptr[y*width+width-2] ] ++;
			ihist_bd[ i_ptr[y*width+width-1] ] ++;
		}

		for( int i=0; i<n_segs; i++ )
		{
			loca_prior[i] *= max(0.5f, (1 - sqrt(ihist_bd[i]/(hist[i]*WH)) * 3));
		}



		///////////////////////////////////////////////////////////////////////////////////////////
		//
		// Step 3.4: calculate region contrast
		//           saliency of region is the sum of its distance to others
		//           weighted by corresponding region size
		//
		///////////////////////////////////////////////////////////////////////////////////////////

		float max_sal = 0, min_sal = 1000000;

		sal[k] = new float[n_segs];
		memset(sal[k], 0, n_segs*sizeof(float));

		u_ptr = clr_rgb.data;
		for( int i=0; i<n_segs; i++ )
		{

			for( int j=0; j<n_segs; j++ )
			{
				sal[k][i] += hist[j] * clr_contrast[i*n_segs+j];
			}

			sal[k][i] *= (0.5f+min(0.25f,hist[i]));

			max_sal = max( max_sal, sal[k][i]);
			min_sal = min( min_sal, sal[k][i]);
		}


		///////////////////////////////////////////////////////////////////////////////////////////
		//
		// Step 3.5: normalize & apply location prior
		//
		///////////////////////////////////////////////////////////////////////////////////////////

		float max_sal2 = 0, min_sal2= 1000000;
		min_sal = min_sal*0.5f;

		if ( n_segs==1 ) 
		{
			sal[k][0] = 255.0f;
		}
		else
		{

			for( int i=0; i<n_segs; i++ )
			{
				sal[k][i] = (sal[k][i]-min_sal) / (max_sal-min_sal) * 255.0f;	
					
				sal[k][i] *= loca_prior[i];//*= (1+scales[k]*scales[k]*loca_prior[i]);


				max_sal2 = max(sal[k][i], max_sal2);
				min_sal2 = min(sal[k][i], min_sal2);
			}
			
			min_sal2 = max(0.0f, min_sal2);
			for( int i=0; i<n_segs; i++ )
			{
				sal[k][i] = (sal[k][i]-min_sal2) / (max_sal2-min_sal2) * 255.0f;	
			}

		}

		///////////////////////////////////////////////////////////////////////////////////////////
		//
		// color smooth (connected similar regions are assigned similar saliency)
		//
		///////////////////////////////////////////////////////////////////////////////////////////

		float w_tmp;
		float sigma_r_color_smooth = 2*40*40;
		float sigma_s_color_smooth = sigma_s_contrast;

		memset( sal_tmp, 0, sizeof(float)*n_segs);
		memset( wt_tmp, 0, sizeof(float)*n_segs);

		for (int si=0; si<n_segs; si++)
		{

			wt_tmp[si] +=  hist_sqrt[si]*2;
			sal_tmp[si] += sal[k][si]*hist_sqrt[si]*2;
			for (int sj=si+1; sj<n_segs; sj++)
			{

				if ( (seg_lists[k])[si]->hasNeighbor(sj) ) 
				{
					w_tmp = exp( - clr_distance[si*n_segs+sj]/(sigma_r_color_smooth)  );
				}
				else
				{
					w_tmp = exp( - clr_distance[si*n_segs+sj]/(sigma_r_color_smooth)  
						- (seg_lists[k])[si]->TransformedDistance((seg_lists[k])[sj], width, height)/(sigma_s_color_smooth) );
				}
				sal_tmp[si] += sal[k][sj] * w_tmp * hist_sqrt[sj];
				sal_tmp[sj] += sal[k][si] * w_tmp * hist_sqrt[si];
				wt_tmp[si] += w_tmp * hist_sqrt[sj];
				wt_tmp[sj] += w_tmp * hist_sqrt[si];

			}

		}

		max_sal2 = 0, min_sal2= 1000000;
		for (int si=0; si<n_segs; si++)
		{
			sal[k][si] = sal_tmp[si]/wt_tmp[si];// printf("%f\n" , sal[k][si]);
			max_sal2 = max(sal[k][si], max_sal2);
			min_sal2 = min(sal[k][si], min_sal2);
		}
		
		if ( n_segs==1 )
		{
			sal[k][0] = 255.0f;
		}
		else 
		{
			SalScale(sal[k], hist, n_segs);
		}



		///////////////////////////////////////////////////////////////////////////////////////////
		//
		// Step 3.6: finally, map saliency of each region to image
		//           region label is stored in 'seg_labels'
		//
		///////////////////////////////////////////////////////////////////////////////////////////

		i_ptr = seg_labels[k].ptr<int>();
		u_ptr = sal_maps[k].ptr<uchar>();
		for (int y=0; y<WH; y++)
		{
			u_ptr[y] = sal[k][ i_ptr[y] ];
		}


		///////////////////////////////////////////////////////////////////////////////////////////
		//
		// Step 3.7: save saliency result
		//
		///////////////////////////////////////////////////////////////////////////////////////////
#ifdef SAVE_LAYER_RESULT

#ifdef IMAGE_DOWNSAMPLE
		sprintf(buffer, "%s%s_RES_%02d.png", out_dir, in_file_no_ext, scales[k]);
		resize(sal_maps[k], tmp_mat, Size(), 2, 2);
		imwrite(buffer, tmp_mat);

		sprintf(buffer, "%s%s_SEG_%02d.png", out_dir, in_file_no_ext, scales[k]);
		resize(seg_ress[k], tmp_mat, Size(ori_width, ori_height));
		imwrite(buffer, tmp_mat);
#else
		sprintf(buffer, "%s%s_RES_%02d.png", out_dir, in_file_no_ext, scales[k]);
		imwrite(buffer, sal_maps[k]);

		sprintf(buffer, "%s%s_SEG_%02d.png", out_dir, in_file_no_ext, scales[k]);
		imwrite(buffer, seg_ress[k]);		
#endif
		
#endif
	}	


	



#ifdef SALIENCY_TIMER_ALL
	end = clock();
	printf("Stage 3: Compute saliency (%.6fs)\n", (double)(end-start)/CLOCKS_PER_SEC);
#endif


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
#ifndef GENERATE_CHS_DATA_FOR_INFERENCE

	///////////////////////////////////////////////////////////////////////////////////////////////
	//
	// Stage 4: Merge (HS model: using a hierarchical model, for detail please refer to our paper)
	//
	///////////////////////////////////////////////////////////////////////////////////////////////

#ifdef SALIENCY_TIMER_ALL
	start = clock();
#endif


	SegTree tree;
	tree.BuildTreeFromSeg(seg_labels, seg_lists, sal, N_SCALES);

	sal_merged = new float [seg_lists[0].size()];
	tree.BPL2T(sal_merged, 255, 0, N_SCALES, lambda, beta);
	

	float min_sal = 10000, max_sal = 0;
	for (int i=0; i<seg_lists[0].size(); i++)
	{
		min_sal = min( min_sal, sal_merged[i]);
		max_sal = max( max_sal, sal_merged[i]);
	}
		
	for (int i=0; i<seg_lists[0].size(); i++)
	{
		sal_merged[i] = ( ( sal_merged[i] - min_sal ) / ( max_sal - min_sal ) * 255.0f + 0.5f);
	}

	u_ptr = sal_maps[0].ptr<uchar>();
	i_ptr = seg_labels[0].ptr<int>();
	for (int y=0; y<WH; y++)
	{
		u_ptr[y] = sal_merged[ i_ptr[y] ];
	}


#ifdef AVERAGE_TIME
	CEND();
#endif

		
	sprintf(buffer, "%s%s_HS.png", out_dir, in_file_no_ext);
#ifdef IMAGE_DOWNSAMPLE
	resize(sal_maps[0], tmp_mat, Size(ori_width, ori_height));
	imwrite(buffer, tmp_mat);
#else
#ifdef REMOVE_IMAGE_FRAME
	Mat salMap_new;
	copyMakeBorder(sal_maps[0], salMap_new, top, bot, lef, rig, BORDER_CONSTANT, 0);
	imwrite(buffer, salMap_new);
#else
	imwrite(buffer, sal_maps[0]);
#endif
#endif



#ifdef SALIENCY_TIMER_ALL
	end = clock();
	printf("Stage 4: Merge saliency (%.6fs)\n", (double)(end-start)/CLOCKS_PER_SEC);
	printf("Total time: %.6fs\n", (double)(end-start0)/CLOCKS_PER_SEC);
#endif

#else
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Use Matlab for inference (CHS model: Local-consistent inference)
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	int border[4] = {0};
#ifdef REMOVE_IMAGE_FRAME
	border[0] = top;
	border[1] = bot;
	border[2] = lef;
	border[3] = rig;
#endif
	sprintf(buffer, "%s%s.mat", out_dir, in_file_no_ext);
	SaveSegList2(buffer, seg_labels, seg_ress, seg_lists, sal, scales, N_SCALES, border);

#ifdef AVERAGE_TIME
	CEND();
#endif

#endif
	
	///////////////////////////////////////////////////////////////////////////////////////////////
	//
	// Clearing
	//
	///////////////////////////////////////////////////////////////////////////////////////////////


	delete []hist;
	delete []ihist_bd;
	delete []clr_distance;
	delete []clr_contrast;
	delete []loca_prior;
	delete []cent_wt;
	delete []size_wt;
	delete []sal_tmp;
	delete []wt_tmp;


	for ( int k=0; k<N_SCALES; k++ )
	{
		delete [](sal[k]);
		delete [](clr_lists_rgb[k]);
		
		ClearSegList(seg_lists[k]);
	}
	
	delete []sal_merged;
}


///////////////////////////////////////////////////////////////////////////////////////////////
//
// Save parameters use in this run with time stamp
//
///////////////////////////////////////////////////////////////////////////////////////////////
void SaveParams(char *out_dir, int argc, char **argv)
{
	char buffer[512];
	sprintf(buffer, "%s%s", out_dir, "param.txt");
	FILE *txt = fopen(buffer, "w");

	if (txt==NULL)
	{
		printf("Can't open file '%s' for writing!\n", buffer);
		return;
	}

	MyTime currentTime;

	fprintf(txt, "%s\n", currentTime.toString(buffer, 512));

	for (int i=0; i<argc; i++)
	{
		fprintf(txt, "%s ", argv[i]);
	}

	fclose(txt);
}



int main(int argc,char* argv[])
{	
	///////////////////////////////////////////////////////////////////////////////////////////////
	//
	// check parameters
	//
	///////////////////////////////////////////////////////////////////////////////////////////////

	if( argc<ARGC_MIN )
	{
		printf("You must specify the input image and output dir\n");
		printf("\nUsage:\n");
		printf("\tHSaliency.exe in_img out_dir scale2 scale3\n");
		printf("\tThe minimum scale is fixed to 3. Default parameters for the other two scales are 17 33.\n");
		printf("\nExample:\n");
		printf("Use default parameters:\n\tHSaliency.exe .\\in.png .\\out_dir\\ \n");
		printf("Use your specified scales:\n\tHSaliency.exe .\\in.png .\\out_dir\\ 17 33 \n");
		return 0;
	}
	

	///////////////////////////////////////////////////////////////////////////////////////////////
	//
	// find input 
	//
	///////////////////////////////////////////////////////////////////////////////////////////////
	int nfiles;
	vector<string> files;
	nfiles = GetFiles(argv[1], files);



	///////////////////////////////////////////////////////////////////////////////////////////////
	//
	// default parameters
	//
	///////////////////////////////////////////////////////////////////////////////////////////////
#ifdef IMAGE_DOWNSAMPLE
	int min_reg[N_SCALES] = {1, 75, 250};
	int scales[N_SCALES] = {3, 9, 17};
	min_reg[0] = scales[0]*scales[0];
	min_reg[1] = scales[1]*scales[1];
	min_reg[2] = scales[2]*scales[2];
#else
	int min_reg[N_SCALES] = {1, 150, 500};	// minimum region size (actually not used)
	int scales[N_SCALES] = {5, 17, 33};		// scale parameters
	min_reg[0] = scales[0]*scales[0];
	min_reg[1] = scales[1]*scales[1];
	min_reg[2] = scales[2]*scales[2];
#endif
	int sigma_s = 7;						// sigma spatial
	float sigma_r = 6.5;					// sigma range
	int noSmooth = 0;						// smooth filtering off
	float bdWeight = 1;						// boundary weight

	float beta[N_SCALES+1] = {0.5, 4, 2, 0.01};
	float lambda[N_SCALES+1] = {4, 4, 0.01, 0.01};


	if (argc>ARGC_SCALES_ALL)
	{
//		scales[0] = atoi(argv[ARGC_SCALES_ALL-2]);
		scales[1] = atoi(argv[ARGC_SCALES_ALL-1]);
		scales[2] = atoi(argv[ARGC_SCALES_ALL]);
//		min_reg[0] = scales[0]*scales[0];
		min_reg[1] = scales[1]*scales[1];
		min_reg[2] = scales[2]*scales[2];
		printf("user selected scales (%d,%d,%d)\n", scales[0], scales[1], scales[2]);
	}
	

	///////////////////////////////////////////////////////////////////////////////////////////////
	//
	// Process
	//
	///////////////////////////////////////////////////////////////////////////////////////////////

	if ( nfiles==0 )
	{
		printf("The specified path is wrong.\n");
		return 0;
	}
	else
	{
		// Save parameters in the output directory
		SaveParams(argv[2], argc, argv);


		// Find file directory
		char buffer[1024];
		int id = strlen(argv[1])-1;

		while ( argv[1][id]!='\\' && argv[1][id]!='/' && id>0 ) id--;
		if ( id>0 )
		{
			strcpy( buffer, argv[1] );
			buffer[id+1] = '\0';
		}
		else
		{
			id = -1;
			buffer[0] = '\0';
		}


#ifdef AVERAGE_TIME
 		CCLEAR();
#endif
//#pragma omp parallel for
		for ( int i=0; i<nfiles; i++ )
		{
			// process each image specified in the input argument

			char *buffer2 = buffer;
			strcpy(buffer2+id+1, files[i].c_str());
			printf("(%d/%d) %s\n", i+1, nfiles, buffer2);

			MultiscaleSaliency(buffer2, argv[2], min_reg, scales, sigma_s, sigma_r, noSmooth, bdWeight, beta, lambda);
			
		}

#ifdef AVERAGE_TIME
		cummulateTime /= nfiles;
		CSUMMARY();
#endif
	}
}

